{% extends "!layout.html" %}
{%- block extrahead %}
{%- if render_sidebar %}
    <!-- Original Alabaster Extends -->
    {{ super() }}
    <link rel="stylesheet" href="{{ pathto('_static/custom.css', 1) }}" type="text/css" />

    <meta name="viewport" content="width=device-width, initial-scale=0.9, maximum-scale=0.9" />
    
    <!-- Social Meta Tag Guides: -->
    <!-- https://metatags.io -->
    <!-- https://css-tricks.com/essential-meta-tags-social-media/ -->

    <!-- Primary Meta Tags -->
    <meta property="title" content="MicroStructPy - Microstructure Mesh Generation in Python">
    <meta property="description" content="Microstructure modeling, mesh generation, analysis, and visualization.">

    <!-- Open Graph / Facebook Meta Tags -->
    <meta property="og:title" content="MicroStructPy - Microstructure Mesh Generation in Python">
    <meta property="og:site_name" content="MicroStructPy">
    <meta property="og:description" content="Microstructure modeling, mesh generation, analysis, and visualization.">
    <meta property="og:image" content="https://docs.microstructpy.org/en/latest/_static/logo/social.png">
    <meta property="og:url" content="https://docs.microstructpy.org">
    <meta property="og:type" content="website">

    <!-- Twitter Meta Tags-->
    <meta property="twitter:card" content="summary_large_image">
    <meta property="twitter:url" content="https://docs.microstructpy.org">
    <meta property="twitter:title" content="MicroStructPy - Microstructure Mesh Generation in Python">
    <meta property="twitter:description" content="Microstructure modeling, mesh generation, analysis, and visualization.">
    <meta property="twitter:image" content="https://docs.microstructpy.org/en/latest/_static/logo/social.png">
    <meta property="twitter:image:alt" content="MicroStructPy social banner">

    <!-- Touch Icons / Favicons Guide -->
    <!-- https://www.emergeinteractive.com/insights/detail/the-essentials-of-favicons/ -->

    <!-- Touch Icons - Generic -->
    <link rel="icon" sizes="228x228" href="_static/logo/logo_228x228.png">
    <link rel="icon" sizes="192x192" href="_static/logo/logo_192x192.png">
    <link rel="icon" sizes="128x128" href="_static/logo/logo_128x128.png">
    <link rel="icon" sizes="96x96" href="_static/logo/logo_96x96.png">
    <link rel="icon" sizes="76x76" href="_static/logo/logo_76x76.png">
    <link rel="icon" sizes="57x57" href="_static/logo/logo_57x57.png">
    <link rel="icon" sizes="32x32" href="_static/logo/logo_32x32.png">

    <!-- Touch Icons - Android -->
    <link rel="shortcut icon" sizes="196x196" href="_static/logo/logo_196x196.png">

    <!-- Touch Icons - iOS-->
    <link rel="apple-touch-icon-precomposed" sizes="180x180" href="_static/logo/logo_180x180.png">
    <link rel="apple-touch-icon-precomposed" sizes="152x152" href="_static/logo/logo_152x152.png">
    <link rel="apple-touch-icon-precomposed" sizes="144x144" href="_static/logo/logo_144x144.png">
    <link rel="apple-touch-icon-precomposed" sizes="120x120" href="_static/logo/logo_120x120.png">
    <link rel="apple-touch-icon-precomposed" sizes="114x114" href="_static/logo/logo_114x114.png">
    <link rel="apple-touch-icon-precomposed" sizes="76x76" href="_static/logo/logo_76x76.png">
    <link rel="apple-touch-icon-precomposed" sizes="72x72" href="_static/logo/logo_72x72.png">
    <link rel="apple-touch-icon-precomposed" href="_static/logo/logo_57x57.png">

    <!-- Touch Icons - Windows 8 IE 10 -->
    <meta name="msapplication-TileColor" content="#FFFFFF">
    <meta name="msapplication-TileImage" content="_static/logo/logo_144x144.png">

    <!-- Touch Icons - Windows 8.1 + IE11 and above -->
    <meta name="msapplication-config" content="_static/logo/browserconfig.xml" />
{% endif %}
{%- endblock %}
